/* A list of (IP, [:addr:]value) pairs:
 *  IP - use default addr and value, subst IP
 *  IP :addr: - use this addr and default value, subst IP
 *  IP value - default addr and given value, subst IP
 *  IP =subst - default addr and value, substitute subst
 *  IP :addr:=subst - this addr and default value, substitute subst
 * A list is stored as a sorted array, with binary search --
 * maybe not the most efficient way, but it's the most optimal
 * from the memory requiriments point of view, simple to
 * handle and sufficient enouth (imho)
 */

#include <netinet/in.h>
#include <arpa/inet.h>
#include <syslog.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "rbldnsd.h"
#include "dns.h"
#include "sstr.h"

struct entry {
  struct in_addr addr;
  struct in_addr a;
  char *txt; /* sstr */
};

struct zonedata {
  unsigned nent;
  struct entry *ent;
  char *txt; /* sstr */
};

static struct zonedata *freezone(struct zonedata *z) {
  if (z) {
    if (z->txt) sstrfree(z->txt);
    if (z->ent) {
      unsigned i;
      for (i = 0; i < z->nent; ++i)
        if (z->ent[i].txt) sstrfree(z->ent[i].txt);
      free(z->ent);
    }
    free(z);
  }
  return NULL;
}

static int cmpent(const void *a, const void *b) {
  if (((struct entry*)a)->addr.s_addr < ((struct entry*)b)->addr.s_addr)
    return -1;
  if (((struct entry*)a)->addr.s_addr > ((struct entry*)b)->addr.s_addr)
    return 1;
  return 0;
}

struct zonedata *ipvset_load(const char *name, struct zonedata *z) {
  int lineno = 0;
  int l;
  struct entry *ent = NULL;
  unsigned nent = 0, nalc = 0;
  struct in_addr k, a, r;
  char *p;
  FILE *f;
  char buf[256];
  freezone(z);
  if ((f = fopen(name, "r")) == NULL) {
    log(LOG_ERR, "unable to open zone file `%s': %s", name, strerror(errno));
    return NULL;
  }
  if ((z = (struct zonedata *)emalloc(sizeof(*z))) == NULL) {
    fclose(f);
    return NULL;
  }
  z->ent = NULL;
  z->txt = NULL;
  r = default_aaddr;

  while(fgets(buf, sizeof(buf), f)) {
    ++lineno;
    l = strlen(buf);
    if (l && buf[l-1] == '\n') buf[--l] = '\0';
    if (!l || buf[0] == '#') continue;
    if (lineno == 1 && buf[0] == ':') {
      if (!addrtxt(buf, &a, &p))
        log(LOG_WARNING, "zone %s line 1: invalid default entry", name);
      else {
        r = a;
	if (*p && (z->txt = esstrdup(p)) == NULL) {
	  fclose(f);
	  return freezone(z);
	}
      }
      continue;
    }
    for(p = buf; *p; ++p)
      if (*p == ' ' || *p == '\t') break;
    if (*p)
      *p++ = '\0';
    if ((l = ip4range(buf, &k)) == 0) {
      log(LOG_WARNING, "zone %s, line %d: invalid address %s",
          name, lineno, buf);
      continue;
    }
    if (!addrtxt(p, &a, &p)) {
      log(LOG_WARNING, "zone %s, line %d: invalid entry", name, lineno);
      continue;
    }
    if (nent + l >= nalc) {
      do nalc = nalc ? nalc * 2 : 512;
      while(nent + l > nalc);
      z->ent = ent = (struct entry *)erealloc(ent, nalc * sizeof(*ent));
      if (!ent) {
	fclose(f);
	return freezone(z);
      }
    }
    while(l--) {
      ent[nent].addr = k;
      ent[nent].a = a.s_addr ? a : r;
      if (!*p)
        ent[nent].txt = NULL;
      else if ((ent[nent].txt = esstrdup(p)) == NULL) {
        fclose(f);
        return freezone(z);
      }
      z->nent = ++nent;
      ++k.s_addr;
    }
  }
  fclose(f);
  z->ent = (struct entry *)realloc(ent, nent * sizeof(*ent));
  qsort(ent, nent, sizeof(*ent), cmpent);
  return z;
}

int ipvset_query(const struct zonedata *const z,
	    struct dnspacket *p,
	    const unsigned char *const query, unsigned qflg)
{
   int a = 0, b = z->nent - 1, m;
   const struct entry *e = z->ent;
   struct in_addr addr;
   if (!dn4octet(query, &addr))
     return 0;
   while(a <= b) {
     m = (a + b) >> 1;
     if (e[m].addr.s_addr == addr.s_addr) {
       a = m; while(a > 0 && e[a-1].addr.s_addr == addr.s_addr) --a;
       b = m + 1; while(b < z->nent && e[b].addr.s_addr == addr.s_addr) ++b;
       if (qflg & QUERY_A)
         for(m = a; m < b; ++m) addrec_a(p, e[m].a);
       if (qflg & QUERY_TXT) {
         const char *txt, *subst;
         for(m = a; m < b; ++m) {
           if (!e[m].txt) {
             if ((txt = z->txt) == NULL)
               continue;
             subst = inet_ntoa(e[m].addr);
           }
           else if (e[m].txt[0] != '=') {
             txt = e[m].txt;
             subst = inet_ntoa(e[m].addr);
           }
           else if ((txt = z->txt) != NULL)
             subst = e[m].txt + 1;
	   else
             txt = e[m].txt + 1, subst = NULL;
           addrec_txt(p, txt, subst);
         }
       }
       return 1;
     }
     else if (e[m].addr.s_addr > addr.s_addr) b = m - 1;
     else a = m + 1;
   }
   return 0;
}
