/* A list of (name, [:addr:]value) pairs:
 *  name - use default addr and value, subst name
 *  name :addr: - use this addr and default value, subst name
 *  name value - default addr and given value, subst name
 *  name =subst - default addr and value, substitute subst
 *  name :addr:=subst - this addr and default value, substitute subst
 * A list is stored as a sorted array, with binary search --
 * maybe not the most efficient way, but it's the most optimal
 * from the memory requiriments point of view, simple to
 * handle and sufficient enouth (imho)
 */

#include <netinet/in.h>
#include <arpa/inet.h>
#include <syslog.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "rbldnsd.h"
#include "dns.h"
#include "mempool.h"
#include "sstr.h"

struct entry {
  unsigned char *dn; /* mpool */
  struct in_addr a;
  char *txt; /* sstr */
};

struct zonedata {
  unsigned nent;
  struct entry *ent;
  char *txt; /* sstr */
  struct mempool mp;
};

static struct zonedata *freezone(struct zonedata *z) {
  if (z) {
    mp_free(&z->mp);
    if (z->txt) sstrfree(z->txt);
    if (z->ent) {
      unsigned i;
      for(i = 0; i < z->nent; ++i)
        if (z->ent[i].txt) sstrfree(z->ent[i].txt);
      free(z->ent);
    }
    free(z);
  }
  return NULL;
}

static int cmpent(const void *a, const void *b) {
  return strcmp(((struct entry*)a)->dn, ((struct entry*)b)->dn);
}

struct zonedata *nvset_load(const char *name, struct zonedata *z) {
  int lineno = 0;
  int l;
  struct entry *ent = NULL;
  unsigned nent = 0, nalc = 0;
  struct in_addr a, r;
  char *p;
  unsigned char dn[DNS_MAXDN+1];
  FILE *f;
  char buf[256];
  freezone(z);
  if ((f = fopen(name, "r")) == NULL) {
    log(LOG_ERR, "unable to open zone file `%s': %s", name, strerror(errno));
    return NULL;
  }
  if ((z = (struct zonedata *)emalloc(sizeof(*z))) == NULL) {
    fclose(f);
    return NULL;
  }
  mp_init(&z->mp);
  z->ent = NULL;
  z->txt = NULL;
  r = default_aaddr;

  while(fgets(buf, sizeof(buf), f)) {
    ++lineno;
    l = strlen(buf);
    if (l && buf[l-1] == '\n') buf[--l] = '\0';
    if (!l || buf[0] == '#') continue;
    if (lineno == 1 && buf[0] == ':') {
      if (!addrtxt(buf, &a, &p))
        log(LOG_WARNING, "zone %s line 1: invalid default entry", name);
      else {
        r = a;
	if (z->txt) sstrfree(z->txt);
	if (!*p)
          z->txt = NULL;
	else if ((z->txt = esstrdup(p)) == NULL) {
	  fclose(f);
	  return freezone(z);
	}
      }
      continue;
    }
    for(p = buf; *p; ++p)
      if (*p == ' ' || *p == '\t') break;
    if (*p)
      *p++ = '\0';
    if (!dns_ptodn(buf, dn, sizeof(dn))) {
      log(LOG_WARNING, "zone %s, line %d: invalid domain name %s",
          name, lineno, buf);
      continue;
    }
    dns_dntol(dn, dn, sizeof(dn));
    if (!addrtxt(p, &a, &p)) {
      log(LOG_WARNING, "zone %s, line %d: invalid entry", name, lineno);
      continue;
    }
    if (nent == nalc) {
      nalc = nalc ? nalc * 2 : 512;
      z->ent = ent = (struct entry *)erealloc(ent, nalc * sizeof(*ent));
      if (!ent) {
	fclose(f);
	return freezone(z);
      }
    }
    ent[nent].a = a.s_addr ? a : r;
    if ((ent[nent].dn = mp_estrdup(&z->mp, dn)) == NULL) {
      fclose(f);
      return freezone(z);
    }
    if (!*p)
      ent[nent].txt = NULL;
    else if ((ent[nent].txt = esstrdup(p)) == NULL) {
      fclose(f);
      return freezone(z);
    }
    z->nent = ++nent;
  }
  fclose(f);
  z->ent = (struct entry *)realloc(ent, nent * sizeof(*ent));
  qsort(ent, nent, sizeof(*ent), cmpent);
  return z;
}

int nvset_query(const struct zonedata *const z,
	    struct dnspacket *p,
	    const unsigned char *const query, unsigned qflg)
{
   int a = 0, b = z->nent - 1, m, r;
   const struct entry *e = z->ent;
   while(a <= b) {
     m = (a + b) >> 1;
     r = strcmp(query, e[m].dn);
     if (!r) {
       a = m; while(a > 0 && strcmp(e[a-1].dn, query) == 0) --a;
       b = m + 1; while(b < z->nent && strcmp(e[b].dn, query) == 0) ++b;
       if (qflg & QUERY_A)
         for(m = a; m < b; ++m) addrec_a(p, e[m].a);
       if (qflg & QUERY_TXT) {
         const char *txt, *subst;
         char dnname[DNS_MAXDOMAIN];
         for(m = a; m < b; ++m) {
           if (!e[m].txt) {
             if ((txt = z->txt) == NULL)
               continue;
             dns_dntop(e[m].dn, dnname, DNS_MAXDOMAIN);
             subst = dnname;
           }
           else if (e[m].txt[0] != '=') {
             txt = e[m].txt;
             dns_dntop(e[m].dn, dnname, DNS_MAXDOMAIN);
             subst = dnname;
           }
           else if ((txt = z->txt) != NULL)
             subst = e[m].txt + 1;
	   else
             txt = e[m].txt + 1, subst = NULL;
           addrec_txt(p, txt, subst);
         }
       }
       return 1;
     }
     else if (r < 0) b = m - 1;
     else a = m + 1;
   }
   return 0;
}
